"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.run = exports.parseOptions = exports.printHelp = exports.OutputFormat = exports.description = exports.name = void 0;
const ballot_encoder_1 = require("@votingworks/ballot-encoder");
const chalk_1 = __importDefault(require("chalk"));
const fs_1 = require("fs");
const table_1 = require("table");
const __1 = require("..");
const __2 = require("../..");
const Interpreter_1 = require("../../Interpreter");
const images_1 = require("../../utils/images");
exports.name = 'interpret';
exports.description = 'Interpret images of ballot pages';
var OutputFormat;
(function (OutputFormat) {
    OutputFormat["JSON"] = "json";
    OutputFormat["Table"] = "table";
})(OutputFormat = exports.OutputFormat || (exports.OutputFormat = {}));
function makeInputFromBallotArgument(arg) {
    const [ballotPath, metadataPath] = arg.split(':');
    const input = {
        id: () => ballotPath,
        imageData: () => images_1.loadImageData(ballotPath),
        metadata: async () => metadataPath
            ? JSON.parse(await fs_1.promises.readFile(metadataPath, 'utf-8'))
            : undefined,
    };
    return input;
}
function printHelp($0, out) {
    out.write(`${$0} interpret -e JSON IMG1 [IMG2 …]\n`);
    out.write(`\n`);
    out.write(chalk_1.default.italic(`Examples\n`));
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Interpret ballots based on a single template.\n`));
    out.write(`${$0} interpret -e election.json -t template.png ballot*.png\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Interpret test mode ballots.\n`));
    out.write(`${$0} interpret -e election.json -T -t template.png ballot*.png\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Interpret ballots to JSON.\n`));
    out.write(`${$0} interpret -e election.json -f json template*.png ballot*.png\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Specify image metadata (file:metdata-file).\n`));
    out.write(`${$0} interpret -e election.json template1.png:template1-metadata.json template2.png:template2-metdata.json ballot1.png:ballot1-metadata.json\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Set an explicit minimum mark score (0-1).\n`));
    out.write(`${$0} interpret -e election.json -m 0.5 template*.png ballot*.png\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Automatically process images as templates until all pages are found.\n`));
    out.write(`${$0} interpret -e election.json image*.png\n`);
}
exports.printHelp = printHelp;
async function parseOptions({ commandArgs: args, }) {
    let election;
    let testMode = false;
    const autoInputs = [];
    const templateInputs = [];
    const ballotInputs = [];
    let markScoreVoteThreshold = Interpreter_1.DEFAULT_MARK_SCORE_VOTE_THRESHOLD;
    let format = OutputFormat.Table;
    for (let i = 0; i < args.length; i += 1) {
        const arg = args[i];
        switch (arg) {
            case '-e':
            case '--election': {
                i += 1;
                const electionJSONFile = args[i];
                if (!electionJSONFile || electionJSONFile.startsWith('-')) {
                    throw new __1.OptionParseError(`Expected election definition file after ${arg}, but got ${electionJSONFile || 'nothing'}.`);
                }
                election = ballot_encoder_1.parseElection(JSON.parse(await fs_1.promises.readFile(electionJSONFile, 'utf8')));
                break;
            }
            case '-m':
            case '--min-mark-score': {
                i += 1;
                const thresholdString = args[i];
                markScoreVoteThreshold =
                    parseFloat(thresholdString) *
                        (thresholdString.endsWith('%') ? 0.01 : 1);
                if (isNaN(markScoreVoteThreshold)) {
                    throw new __1.OptionParseError(`Invalid minimum mark score: ${args[i]}`);
                }
                break;
            }
            case '-f':
            case '--format':
                i += 1;
                format = args[i].toLowerCase();
                if (format !== OutputFormat.Table && format !== OutputFormat.JSON) {
                    throw new __1.OptionParseError(`Unknown output format: ${format}`);
                }
                break;
            case '-t':
            case '--template': {
                i += 1;
                const templateFile = args[i];
                if (!templateFile || templateFile.startsWith('-')) {
                    throw new __1.OptionParseError(`Expected template file after ${arg}, but got ${templateFile || 'nothing'}`);
                }
                templateInputs.push(makeInputFromBallotArgument(templateFile));
                break;
            }
            case '-b':
            case '--ballot': {
                i += 1;
                const ballotFile = args[i];
                if (!ballotFile || ballotFile.startsWith('-')) {
                    throw new __1.OptionParseError(`Expected ballot file after ${arg}, but got ${ballotFile || 'nothing'}`);
                }
                ballotInputs.push(makeInputFromBallotArgument(ballotFile));
                break;
            }
            case '-T':
            case '--test-mode':
            case '--no-test-mode': {
                testMode = arg !== '--no-test-mode';
                break;
            }
            default: {
                if (arg.startsWith('-')) {
                    throw new __1.OptionParseError(`Unknown option: ${arg}`);
                }
                autoInputs.push(makeInputFromBallotArgument(arg));
                break;
            }
        }
    }
    if (!election) {
        throw new __1.OptionParseError(`Required option 'election' is missing.`);
    }
    return {
        election,
        testMode,
        autoInputs,
        templateInputs,
        ballotInputs,
        markScoreVoteThreshold,
        format,
    };
}
exports.parseOptions = parseOptions;
async function run(options, _stdin, stdout) {
    var _a, _b, _c;
    const interpreter = new __2.Interpreter({
        election: options.election,
        testMode: options.testMode,
        markScoreVoteThreshold: options.markScoreVoteThreshold,
    });
    const ballotInputs = [...options.ballotInputs];
    for (const templateInput of options.templateInputs) {
        await interpreter.addTemplate(await templateInput.imageData(), await ((_a = templateInput.metadata) === null || _a === void 0 ? void 0 : _a.call(templateInput)));
    }
    for (const autoInput of options.autoInputs) {
        const metadata = await ((_b = autoInput.metadata) === null || _b === void 0 ? void 0 : _b.call(autoInput));
        if (!(metadata && interpreter.canScanBallot(metadata))) {
            await interpreter.addTemplate(await autoInput.imageData(), metadata);
        }
        else {
            ballotInputs.push(autoInput);
        }
    }
    const results = [];
    for (const ballotInput of ballotInputs) {
        results.push({
            input: ballotInput,
            interpreted: await interpreter.interpretBallot(await ballotInput.imageData(), await ((_c = ballotInput.metadata) === null || _c === void 0 ? void 0 : _c.call(ballotInput))),
        });
    }
    switch (options.format) {
        case OutputFormat.JSON:
            stdout.write(JSON.stringify(results.map(({ input, interpreted }) => ({
                input: input.id(),
                interpreted: {
                    metadata: interpreted.metadata,
                    votes: interpreted.ballot.votes,
                    marks: interpreted.marks
                        .filter((mark) => mark.type === 'candidate' || mark.type === 'yesno'
                        ? mark.score > 0
                        : true)
                        .map((mark) => {
                        var _a, _b;
                        return mark.type === 'candidate'
                            ? {
                                type: mark.type,
                                contest: mark.contest.id,
                                option: mark.option.id,
                                score: mark.score,
                                bounds: mark.bounds,
                                target: mark.target,
                            }
                            : mark.type === 'yesno'
                                ? {
                                    type: mark.type,
                                    contest: mark.contest.id,
                                    option: mark.option,
                                    score: mark.score,
                                    bounds: mark.bounds,
                                    target: mark.target,
                                }
                                : {
                                    type: mark.type,
                                    contest: (_a = mark.contest) === null || _a === void 0 ? void 0 : _a.id,
                                    option: typeof mark.option === 'string'
                                        ? mark.option
                                        : (_b = mark.option) === null || _b === void 0 ? void 0 : _b.id,
                                    bounds: mark.bounds,
                                };
                    }),
                },
            })), undefined, 2));
            break;
        case OutputFormat.Table:
            stdout.write(table_1.table([
                [
                    chalk_1.default.bold('Contest'),
                    ...results.map(({ input }) => chalk_1.default.bold(input.id())),
                ],
                ...options.election.contests.map((contest) => [
                    contest.title,
                    ...results.map(({ interpreted }) => {
                        var _a, _b;
                        const vote = interpreted.ballot.votes[contest.id];
                        if (contest.type === 'candidate') {
                            const candidates = vote;
                            return ((_a = candidates === null || candidates === void 0 ? void 0 : candidates.map(({ name, isWriteIn }) => isWriteIn ? chalk_1.default.italic(name) : name).join(', ')) !== null && _a !== void 0 ? _a : '');
                        }
                        else {
                            const yesnos = vote;
                            return ((_b = yesnos === null || yesnos === void 0 ? void 0 : yesnos.map((v) => v === 'yes'
                                ? chalk_1.default.green(v)
                                : v === 'no'
                                    ? chalk_1.default.red(v)
                                    : '').join(', ')) !== null && _b !== void 0 ? _b : '');
                        }
                    }),
                ]),
            ]));
            break;
    }
    return 0;
}
exports.run = run;
