"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.run = exports.parseOptions = exports.printHelp = exports.description = exports.name = void 0;
const assert_1 = require("assert");
const chalk_1 = __importDefault(require("chalk"));
const __1 = require("..");
const utils_1 = require("../../../test/utils");
const findContests_1 = __importDefault(require("../../hmpb/findContests"));
const binarize_1 = require("../../utils/binarize");
const imageFormatUtils_1 = require("../../utils/imageFormatUtils");
const path_1 = require("../../utils/path");
const images_1 = require("../../utils/images");
const RGBA_CHANNELS = 4;
const RED_OVERLAY_COLOR = [0xff, 0, 0, 0x60];
const GREEN_OVERLAY_COLOR = [0, 0xff, 0, 0x60];
exports.name = 'layout';
exports.description = 'Annotate the interpreted layout of a ballot page';
function printHelp($0, out) {
    out.write(`${$0} layout IMG1 [IMG2 …]\n`);
    out.write(`\n`);
    out.write(chalk_1.default.italic(`Examples\n`));
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Annotate layout for a single ballot page.\n`));
    out.write(`${$0} layout ballot01.png\n`);
    out.write(`\n`);
    out.write(chalk_1.default.gray(`# Annotate layout for many ballot pages.\n`));
    out.write(`${$0} layout ballot*.png\n`);
}
exports.printHelp = printHelp;
async function parseOptions({ commandArgs: args, }) {
    const ballotImagePaths = [];
    for (const arg of args) {
        if (arg.startsWith('-')) {
            throw new __1.OptionParseError(`unexpected option passed to 'layout': ${arg}`);
        }
        ballotImagePaths.push(arg);
    }
    return { ballotImagePaths };
}
exports.parseOptions = parseOptions;
/**
 * Finds features in an image and writes an image adjacent with overlays marking
 * those features.
 */
async function run(options, _stdin, stdout) {
    for (const ballotImagePath of options.ballotImagePaths) {
        const imageData = await images_1.loadImageData(ballotImagePath);
        const binarized = {
            data: new Uint8ClampedArray(imageData.data.length),
            width: imageData.width,
            height: imageData.height,
        };
        binarize_1.binarize(imageData, binarized);
        const threeColumnContests = [
            ...findContests_1.default(imageData, { columns: [true, true, true] }),
        ];
        const contests = threeColumnContests.length === 0
            ? [...findContests_1.default(imageData, { columns: [true, true] })]
            : threeColumnContests;
        const targetWidth = Math.max(15, Math.round(imageData.width * 0.01));
        for (const contest of contests) {
            fill(imageData, contest.bounds, GREEN_OVERLAY_COLOR);
            for (const corner of contest.corners) {
                drawTarget(imageData, corner, RED_OVERLAY_COLOR, targetWidth);
            }
        }
        const layoutFilePath = path_1.adjacentFile('-layout', ballotImagePath);
        stdout.write(`📝 ${layoutFilePath} ${chalk_1.default.gray(`(${contests.length} contest(s))`)}\n`);
        await utils_1.writeImageToFile(imageData, layoutFilePath);
    }
    return 0;
}
exports.run = run;
/**
 * Draws a target composed of concentric squares around a given point. If the
 * color has transparency, the fill blends with the existing image.
 */
function drawTarget({ data, width, height }, { x, y }, color, size) {
    assert_1.strict.equal(imageFormatUtils_1.getImageChannelCount({ data, width, height }), RGBA_CHANNELS);
    const halfSize = Math.ceil(size / 2);
    for (let xd = -halfSize; xd <= halfSize; xd++) {
        for (let yd = -halfSize; yd <= halfSize; yd++) {
            if ((xd % 2 !== 0 && Math.abs(yd) <= Math.abs(xd)) ||
                (yd % 2 !== 0 && Math.abs(xd) <= Math.abs(yd))) {
                const offset = ((y + yd) * width + (x + xd)) * RGBA_CHANNELS;
                const dst = data.slice(offset, offset + RGBA_CHANNELS);
                data.set(alphaBlend(dst, color), offset);
            }
        }
    }
}
/**
 * Fills a region of an image with a particular color. If the color has
 * transparency, the fill blends with the existing image.
 */
function fill({ data, width, height }, bounds, color) {
    assert_1.strict.equal(imageFormatUtils_1.getImageChannelCount({ data, width, height }), RGBA_CHANNELS);
    for (let y = bounds.y; y < bounds.y + bounds.height; y++) {
        for (let x = bounds.x; x < bounds.x + bounds.width; x++) {
            const offset = (y * width + x) * RGBA_CHANNELS;
            const dst = data.slice(offset, offset + RGBA_CHANNELS);
            data.set(alphaBlend(dst, color), offset);
        }
    }
}
/**
 * Computes the color of a pixel by blending `src` on top of `dst`.
 *
 * @see https://en.wikipedia.org/wiki/Alpha_compositing#Alpha_blending
 */
function alphaBlend(dst, src) {
    const dstR = dst[0];
    const dstG = dst[1];
    const dstB = dst[2];
    const dstA = dst[3];
    const srcR = src[0];
    const srcG = src[1];
    const srcB = src[2];
    const srcA = src[3];
    return [
        (srcR * srcA) / 0xff + ((dstR * dstA) / 0xff) * (1 - srcA / 0xff),
        (srcG * srcA) / 0xff + ((dstG * dstA) / 0xff) * (1 - srcA / 0xff),
        (srcB * srcA) / 0xff + ((dstB * dstA) / 0xff) * (1 - srcA / 0xff),
        (srcA / 0xff + (1 - srcA / 0xff)) * 0xff,
    ];
}
