"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.findBallotLayoutCorrespondance = void 0;
const debug_1 = __importDefault(require("debug"));
const binarize_1 = require("../utils/binarize");
const corners_1 = require("../utils/corners");
const geometry_1 = require("../utils/geometry");
const imageFormatUtils_1 = require("../utils/imageFormatUtils");
const iterators_1 = require("../utils/iterators");
const VisitedPoints_1 = require("../utils/VisitedPoints");
const shapes_1 = require("./shapes");
const debug = debug_1.default('hmpb-interpreter:findContests');
function* findContests(ballotImage, { inset = Math.round(0.035 * ballotImage.width), separation = Math.round(0.0175 * ballotImage.width), columns = [true, true, true], expectedWidth = Math.floor((ballotImage.width - 2 * inset - (columns.length - 1) * separation) /
    columns.length), minExpectedHeight = Math.floor(0.1 * ballotImage.height), maxExpectedHeight = Math.ceil(0.9 * ballotImage.height), errorMargin = Math.ceil(0.025 * ballotImage.width), } = {}) {
    const visitedPoints = new VisitedPoints_1.VisitedPoints(ballotImage.width, ballotImage.height);
    for (const [columnIndex, column] of columns.entries()) {
        if (!column) {
            continue;
        }
        const columnMidX = Math.round(inset + columnIndex * (expectedWidth + separation) + expectedWidth / 2);
        let lastShape;
        const expectedContestTop = findTopBorderInset(ballotImage, columnMidX, {
            yMax: inset - errorMargin,
        }) + inset;
        for (let y = expectedContestTop - errorMargin; y < ballotImage.height - inset - minExpectedHeight + errorMargin; y++) {
            if (!lastShape && y > expectedContestTop + errorMargin) {
                debug('abandoning column %d because no top contest was found by y=%d', columnIndex, y);
                break;
            }
            if (lastShape &&
                y >
                    lastShape.bounds.y +
                        lastShape.bounds.height +
                        separation +
                        errorMargin) {
                debug('abandoning the rest of column %d because we should have found another contest box by y=%d', columnIndex, y);
                break;
            }
            const shape = shapes_1.findShape(ballotImage, { x: columnMidX, y }, visitedPoints);
            if (shape.bounds.width <= 1 || shape.bounds.height <= 1) {
                continue;
            }
            const corners = corners_1.getCorners(shape);
            const rectangle = shapes_1.parseRectangle(corners);
            const cornerBasedWidth = Math.max(geometry_1.euclideanDistance(corners[0], corners[1]), geometry_1.euclideanDistance(corners[2], corners[3]));
            const cornerBasedHeight = Math.max(geometry_1.euclideanDistance(corners[0], corners[2]), geometry_1.euclideanDistance(corners[1], corners[3]));
            if (!rectangle.isRectangle) {
                debug('skipping shape because it is not rectangular: bounds=%O, corners=%O, angles=%O', shape.bounds, corners, rectangle.angles.map((angle) => `${Math.round(((angle * 180) / Math.PI) * 100) / 100}°`));
            }
            else if (cornerBasedHeight < minExpectedHeight ||
                cornerBasedHeight > maxExpectedHeight ||
                cornerBasedWidth < expectedWidth - errorMargin ||
                cornerBasedWidth > expectedWidth + errorMargin) {
                debug('skipping shape because it is the wrong size: bounds=%O, actual=%dˣ%d, min=%dˣ%d, max=%dˣ%d', shape.bounds, cornerBasedWidth, cornerBasedHeight, expectedWidth - errorMargin, minExpectedHeight, expectedWidth + errorMargin, maxExpectedHeight);
            }
            else {
                debug('found contest shape: %O', shape.bounds);
                yield {
                    bounds: shape.bounds,
                    corners,
                };
                lastShape = shape;
            }
            y = shape.bounds.y + shape.bounds.height;
        }
    }
}
exports.default = findContests;
function findTopBorderInset({ data, width, height }, x, { yMax = height - 1, minimumConsecutiveWhitePixels = Math.ceil(height * 0.005), } = {}) {
    debug('looking for top inset at x=%d within %dpx of the top with a run of %d white pixels', x, yMax + 1, minimumConsecutiveWhitePixels);
    const channels = imageFormatUtils_1.getImageChannelCount({ data, width, height });
    // Look for black border within [0, yMax].
    let seen = false;
    let y = 0;
    while (y <= yMax) {
        const color = data[(y * width + x) * channels];
        if (color === binarize_1.PIXEL_BLACK) {
            seen = true;
            break;
        }
        y++;
    }
    if (!seen) {
        // Didn't find one.
        debug('no border found by x=%d y=%d', x, y);
        return 0;
    }
    // Look for a run of white pixels that marks the end of the border.
    let consecutiveWhitePixels = 0;
    while (consecutiveWhitePixels < minimumConsecutiveWhitePixels && y < height) {
        const color = data[(y * width + x) * channels];
        if (color === binarize_1.PIXEL_BLACK) {
            consecutiveWhitePixels = 0;
        }
        else {
            consecutiveWhitePixels++;
            debug('found a white pixel at x=%d y=%d, count=%d', x, y, consecutiveWhitePixels);
        }
        y++;
    }
    if (consecutiveWhitePixels < minimumConsecutiveWhitePixels) {
        debug('did not find the end of a border');
        return 0;
    }
    debug('end of the border found starting at x=%d y=%d', x, y - consecutiveWhitePixels);
    return y - consecutiveWhitePixels;
}
function findBallotLayoutCorrespondance(contests, ballot, template, { allowedScaleErrorRatio = 0.1 } = {}) {
    const expectedAreaScale = (ballot.ballotImage.imageData.width * ballot.ballotImage.imageData.height) /
        (template.ballotImage.imageData.width *
            template.ballotImage.imageData.height);
    const minAreaScale = expectedAreaScale * (1 - allowedScaleErrorRatio);
    const maxAreaScale = expectedAreaScale * (1 + allowedScaleErrorRatio);
    const mismatchedContests = [];
    for (const [definition, templateContest, ballotContest] of iterators_1.zip(contests, template.contests, ballot.contests)) {
        const templateArea = geometry_1.poly4Area(templateContest.corners);
        const ballotArea = geometry_1.poly4Area(ballotContest.corners);
        const areaScale = ballotArea / templateArea;
        if (areaScale < minAreaScale || areaScale > maxAreaScale) {
            mismatchedContests.push({
                definition,
                template: templateContest,
                ballot: ballotContest,
            });
        }
    }
    return {
        corresponds: mismatchedContests.length === 0,
        mismatchedContests,
    };
}
exports.findBallotLayoutCorrespondance = findBallotLayoutCorrespondance;
