"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.parseRectangle = exports.findShape = void 0;
const binarize_1 = require("../utils/binarize");
const geometry_1 = require("../utils/geometry");
const imageFormatUtils_1 = require("../utils/imageFormatUtils");
const VisitedPoints_1 = require("../utils/VisitedPoints");
/**
 * Finds a shape in a binarized image by looking for adjacent pixels of a
 * specific color starting at a given point.
 */
function findShape(imageData, startingPoint, visitedPoints = new VisitedPoints_1.VisitedPoints(imageData.width, imageData.height), { color = binarize_1.PIXEL_BLACK } = {}) {
    const toVisit = [startingPoint];
    const points = new VisitedPoints_1.VisitedPoints(imageData.width, imageData.height);
    const { data, width, height } = imageData;
    const channel = imageFormatUtils_1.getImageChannelCount(imageData);
    const topEdge = new Int32Array(imageData.width).fill(imageData.height);
    const rightEdge = new Int32Array(imageData.height).fill(-1);
    const bottomEdge = new Int32Array(imageData.width).fill(-1);
    const leftBorder = new Int32Array(imageData.height).fill(imageData.width);
    let xMin = startingPoint.x;
    let yMin = startingPoint.y;
    let xMax = startingPoint.x;
    let yMax = startingPoint.y;
    for (let point; (point = toVisit.shift()); point) {
        const { x, y } = point;
        if (!visitedPoints.add(x, y)) {
            continue;
        }
        const index = (x + y * width) * channel;
        const isForeground = data[index] === color;
        points.add(x, y, isForeground);
        if (isForeground) {
            if (x < xMin) {
                xMin = x;
            }
            if (y < yMin) {
                yMin = y;
            }
            if (x > xMax) {
                xMax = x;
            }
            if (y > yMax) {
                yMax = y;
            }
            if (x < leftBorder[y]) {
                leftBorder[y] = x;
            }
            if (y < topEdge[x]) {
                topEdge[x] = y;
            }
            if (x > rightEdge[y]) {
                rightEdge[y] = x;
            }
            if (y > bottomEdge[x]) {
                bottomEdge[x] = y;
            }
            for (const xD of [-1, 0, 1]) {
                for (const yD of [-1, 0, 1]) {
                    const nextX = x + xD;
                    const nextY = y + yD;
                    if (nextX > 0 &&
                        nextY > 0 &&
                        nextX < width &&
                        nextY < height &&
                        !points.has(nextX, nextY)) {
                        toVisit.push({ x: nextX, y: nextY });
                    }
                }
            }
        }
    }
    return {
        bounds: {
            x: xMin,
            y: yMin,
            width: xMax - xMin + 1,
            height: yMax - yMin + 1,
        },
        edges: {
            top: topEdge,
            right: rightEdge,
            bottom: bottomEdge,
            left: leftBorder,
        },
    };
}
exports.findShape = findShape;
/**
 * Determines whether the given corners make for a roughly rectangular shape.
 * The amount of allowed error can be controlled.
 *
 * @param corners the corners to check
 * @param param1.allowedErrorAngle the angle in radians to allow in error
 */
function parseRectangle(corners, { allowedErrorAngle = (5 / 180) * Math.PI } = {}) {
    const [topLeft, topRight, bottomLeft, bottomRight] = corners;
    const minAllowedRightAngle = Math.PI / 2 - allowedErrorAngle;
    const maxAllowedRightAngle = Math.PI / 2 + allowedErrorAngle;
    const topLeftAngle = geometry_1.angleBetweenPoints(bottomLeft, topLeft, topRight);
    const topRightAngle = geometry_1.angleBetweenPoints(topLeft, topRight, bottomRight);
    const bottomLeftAngle = geometry_1.angleBetweenPoints(bottomRight, bottomLeft, topLeft);
    const bottomRightAngle = geometry_1.angleBetweenPoints(topRight, bottomRight, bottomLeft);
    return {
        isRectangle: topLeftAngle >= minAllowedRightAngle &&
            topLeftAngle <= maxAllowedRightAngle &&
            topRightAngle >= minAllowedRightAngle &&
            topRightAngle <= maxAllowedRightAngle &&
            bottomLeftAngle >= minAllowedRightAngle &&
            bottomLeftAngle <= maxAllowedRightAngle &&
            bottomRightAngle >= minAllowedRightAngle &&
            bottomRightAngle <= maxAllowedRightAngle,
        angles: [topLeftAngle, topRightAngle, bottomLeftAngle, bottomRightAngle],
    };
}
exports.parseRectangle = parseRectangle;
