"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.poly4Area = exports.triangleArea = exports.angleBetweenPoints = exports.median = exports.euclideanDistance = exports.editDistance = exports.flipRectVH = exports.roundPoint = exports.rectCenter = exports.rectCorners = void 0;
const assert_1 = require("assert");
/**
 * Gets the four extreme points of a rectangle, inclusive.
 */
function rectCorners({ x, y, width, height }) {
    return [
        { x, y },
        { x: x + width - 1, y },
        { x, y: y + height - 1 },
        { x: x + width - 1, y: y + height - 1 },
    ];
}
exports.rectCorners = rectCorners;
/**
 * Gets the center point of a rectangle, optionally rounded.
 */
function rectCenter({ x, y, width, height }, { round = false } = {}) {
    const center = { x: x + (width - 1) / 2, y: y + (height - 1) / 2 };
    const result = round ? roundPoint(center) : center;
    return result;
}
exports.rectCenter = rectCenter;
/**
 * Rounds a point to the nearest integer axis values.
 */
function roundPoint({ x, y }, { round = Math.round } = {}) {
    return { x: round(x), y: round(y) };
}
exports.roundPoint = roundPoint;
/**
 * Flips a `Rect` vertically and horizontally within another `Rect`, equivalent
 * to a 180° rotation.
 *
 * @param outer bounding rectangle containing `inner` to flip within
 * @param inner rectangle to flip within `outer`
 */
function flipRectVH(outer, inner) {
    return {
        x: outer.x + outer.width - (inner.x + inner.width),
        y: outer.y + outer.height - (inner.y + inner.height),
        width: inner.width,
        height: inner.height,
    };
}
exports.flipRectVH = flipRectVH;
/**
 * Find how many pixel moves it takes to get from a to b.
 */
function editDistance(a, b) {
    return Math.abs(b.x - a.x) + Math.abs(b.y - a.y);
}
exports.editDistance = editDistance;
function euclideanDistance(a, b) {
    return ((b.x - a.x) ** 2 + (b.y - a.y) ** 2) ** 0.5;
}
exports.euclideanDistance = euclideanDistance;
/**
 * Find the median of a list of numbers.
 */
function median(numbers) {
    assert_1.strict(numbers.length > 0);
    if (numbers.length === 1) {
        return numbers[0];
    }
    const sorted = Array.from(numbers).sort();
    if (sorted.length % 2 === 0) {
        const halfway = sorted.length / 2;
        return (sorted[halfway] + sorted[halfway + 1]) / 2;
    }
    return sorted[Math.ceil(sorted.length / 2)];
}
exports.median = median;
/**
 * Compute the inner angle formed by vectors `ba` and `bc`.
 *
 * @see https://math.stackexchange.com/a/361419
 */
function angleBetweenPoints(a, b, c) {
    const abDotBc = (a.x - b.x) * (c.x - b.x) + (a.y - b.y) * (c.y - b.y);
    const abDist = euclideanDistance(a, b);
    const bcDist = euclideanDistance(b, c);
    return Math.acos(abDotBc / (abDist * bcDist));
}
exports.angleBetweenPoints = angleBetweenPoints;
/**
 * Compute the area of a triangle.
 *
 * @see https://www.mathopenref.com/coordtrianglearea.html
 */
function triangleArea(a, b, c) {
    return Math.abs((a.x * (b.y - c.y) + b.x * (c.y - a.y) + c.x * (a.y - b.y)) / 2);
}
exports.triangleArea = triangleArea;
/**
 * Compute the area of a 4-sided polygon.
 */
function poly4Area([topLeft, topRight, bottomLeft, bottomRight,]) {
    return (triangleArea(topLeft, bottomLeft, bottomRight) +
        triangleArea(topLeft, topRight, bottomRight));
}
exports.poly4Area = poly4Area;
