"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.fromRGBA = exports.fromGray = void 0;
const imageFormatUtils_1 = require("./imageFormatUtils");
exports.default = imageFormatUtils_1.makeInPlaceImageTransform(fromGray, fromRGBA);
/**
 * Copies a grayscale image to a destination.
 */
function fromGray(srcImageData, dstImageData = srcImageData) {
    imageFormatUtils_1.assertGrayscaleImage(srcImageData);
    imageFormatUtils_1.assertGrayscaleImage(dstImageData);
    imageFormatUtils_1.assertSizesMatch(srcImageData, dstImageData);
    if (srcImageData === dstImageData) {
        return;
    }
    dstImageData.data.set(srcImageData.data);
}
exports.fromGray = fromGray;
/**
 * Converts an RGBA image to grayscale. If the destination image is a
 * single-channel image, the alpha channel is ignored.
 *
 * Operates on the image in-place by default, or a different destination image
 * may be provided.
 */
function fromRGBA(srcImageData, dstImageData = srcImageData) {
    imageFormatUtils_1.assertRGBAImage(srcImageData);
    imageFormatUtils_1.assertSizesMatch(srcImageData, dstImageData);
    const src32 = new Int32Array(srcImageData.data.buffer);
    if (imageFormatUtils_1.isRGBA(dstImageData)) {
        const dst32 = new Int32Array(dstImageData.data.buffer);
        for (let offset = 0, size = src32.length; offset < size; offset++) {
            const px = src32[offset];
            const r = px & 0xff;
            const g = (px >>> 8) & 0xff;
            const b = (px >>> 16) & 0xff;
            const a = (px >>> 24) & 0xff;
            // Luminosity grayscale formula.
            const luminosity = (0.21 * r + 0.72 * g + 0.07 * b) | 0;
            dst32[offset] =
                luminosity | (luminosity << 8) | (luminosity << 16) | (a << 24);
        }
    }
    else {
        const dst8 = dstImageData.data;
        for (let offset = 0, size = src32.length; offset < size; offset++) {
            const px = src32[offset];
            const r = px & 0xff;
            const g = (px >>> 8) & 0xff;
            const b = (px >>> 16) & 0xff;
            // Luminosity grayscale formula.
            const luminosity = (0.21 * r + 0.72 * g + 0.07 * b) | 0;
            dst8[offset] = luminosity;
        }
    }
}
exports.fromRGBA = fromRGBA;
