"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.countPixels = exports.ratio = void 0;
const binarize_1 = require("../binarize");
const imageFormatUtils_1 = require("../imageFormatUtils");
/**
 * Generates an image from two binarized images where black pixels are where
 * `compare` is black and `base` is not. This is useful for determining where a
 * white-background form was filled out, for example.
 *
 * Note that the sizes of the bounds, which default to the full image size, must
 * be equal.
 *
 * ```
 *         BASE                  COMPARE                 DIFF
 * ┌───────────────────┐  ┌───────────────────┐  ┌───────────────────┐
 * │                   │  │        █ █ ███    │  │        █ █ ███    │
 * │ █ █               │  │ █ █    ███  █     │  │        ███  █     │
 * │  █                │  │  █     █ █ ███    │  │        █ █ ███    │
 * │ █ █ █████████████ │  │ █ █ █████████████ │  │                   │
 * └───────────────────┘  └───────────────────┘  └───────────────────┘
 * ```
 */
function diff(base, compare, baseBounds = { x: 0, y: 0, width: base.width, height: base.height }, compareBounds = {
    x: 0,
    y: 0,
    width: compare.width,
    height: compare.height,
}) {
    imageFormatUtils_1.assertImageChannelsMatch(base, compare);
    imageFormatUtils_1.assertSizesMatch(baseBounds, compareBounds);
    const { data: baseData, width: baseWidth } = base;
    const { data: compareData, width: compareWidth } = compare;
    const { x: baseXOffset, y: baseYOffset } = baseBounds;
    const { x: compareXOffset, y: compareYOffset } = compareBounds;
    const { width: dstWidth, height: dstHeight } = baseBounds;
    let dst;
    if (imageFormatUtils_1.isRGBA(base)) {
        dst = new Uint8ClampedArray(dstWidth * dstHeight * 4);
        for (let y = 0; y < dstHeight; y += 1) {
            for (let x = 0; x < dstWidth; x += 1) {
                const baseOffset = (baseXOffset + x + (baseYOffset + y) * baseWidth) << 2;
                const compareOffset = (compareXOffset + x + (compareYOffset + y) * compareWidth) << 2;
                const dstOffset = (x + y * dstWidth) << 2;
                if (baseData[baseOffset] === binarize_1.PIXEL_BLACK &&
                    baseData[baseOffset + 1] === binarize_1.PIXEL_BLACK &&
                    baseData[baseOffset + 2] === binarize_1.PIXEL_BLACK) {
                    dst[dstOffset] = binarize_1.PIXEL_WHITE;
                    dst[dstOffset + 1] = binarize_1.PIXEL_WHITE;
                    dst[dstOffset + 2] = binarize_1.PIXEL_WHITE;
                }
                else {
                    dst[dstOffset] = compareData[compareOffset];
                    dst[dstOffset + 1] = compareData[compareOffset + 1];
                    dst[dstOffset + 2] = compareData[compareOffset + 2];
                }
                dst[dstOffset + 3] = 0xff;
            }
        }
    }
    else {
        dst = new Uint8ClampedArray(dstWidth * dstHeight);
        for (let y = 0; y < dstHeight; y += 1) {
            for (let x = 0; x < dstWidth; x += 1) {
                const baseOffset = baseXOffset + x + (baseYOffset + y) * baseWidth;
                const compareOffset = compareXOffset + x + (compareYOffset + y) * compareWidth;
                const dstOffset = x + y * dstWidth;
                if (baseData[baseOffset] === binarize_1.PIXEL_BLACK) {
                    dst[dstOffset] = binarize_1.PIXEL_WHITE;
                }
                else {
                    dst[dstOffset] = compareData[compareOffset];
                }
            }
        }
    }
    return {
        data: dst,
        width: dstWidth,
        height: dstHeight,
    };
}
exports.default = diff;
/**
 * Determines the ratio of black (or custom color) pixels in an image to the
 * total number of pixels.
 */
function ratio(image, options = {}) {
    const { width, height } = image;
    return countPixels(image, options) / (width * height);
}
exports.ratio = ratio;
/**
 * Determines number of black (or custom color) pixels in an image.
 */
function countPixels(image, { color = binarize_1.PIXEL_BLACK, bounds = { x: 0, y: 0, width: image.width, height: image.height }, } = {}) {
    const { data } = image;
    const channels = imageFormatUtils_1.getImageChannelCount(image);
    const { x: startX, y: startY, width, height } = bounds;
    let count = 0;
    for (let y = 0; y < height; y += 1) {
        for (let x = 0; x < width; x += 1) {
            if (data[((startY + y) * width + startX + x) * channels] === color) {
                count += 1;
            }
        }
    }
    return count;
}
exports.countPixels = countPixels;
